/*
This MemBrain script has been specifically designed to work with time series in combination with both time invariant and 
time variant nets. 
It uses an approach of preparing the time series input data in a way so that the inputs of the net already contain 
a number of past values in a kind of 'ring buffer' way. Through this approach the net itself can be kept time invariant 
since the data patterns themselves already contain the past values of the time series which shall be taken into account. 

Nevertheless, the script can also be used with time variant nets so that a combination of input values from the past as 
well as time variant net architecture can be used.

What the script does:

- Read in time series data from a CSV file generated by the script user.
- Automatically generate a MemBrain lesson from the time series data which has the data points of time t+1 as expected output to patterns from time point t, 
  t-1, t-2, ... and so on.
  I.e. a lesson which is suitable to train nets which shall try to learn the transfer functions between X(t), X(t-1), Y(t-2),... Y(t), Y(t-1), Y(t-2),... --> X(t+1), Y(t+1)...
- Automatically split the generated lesson into a training and a validation lesson based on a percentage value. Note that this function does not use 
  the built-in lesson split functionality of MemBrain since it may not use random split. Rather than that the validation lesson must be the strict successor 
  of the train lesson in time since the net shall learn time dependent data (i.e. data with implicit order-of-patterns rules).  
- Train an adjustable number of different net candidates with these data for a specified training time and determine the best candidate with respect to the 
  net error based on the validation lesson
- Re-train the best net candidate with the combined lesson (training + validation)
- Record the final answer of the net on this lesson plus an adjustable number of future extrapolated data points
- Export the recorded data into a CSV file including an auto-generated time column to the left of the CSV file

The script in principal supports nets of the following architecture with respect their inputs and outputs. You have to design your nets in this way, too, in 
order to be compatible with the script:

INPUTS
I_TS1(t-n),I_TS1(t-n+1),...,I_TS1(t),I_TS2(t-m),I_TS2(t-m+1),I_TS2(t),I_ADD1,I_ADD2,...,I_ADDN

OUTPUTS
O_TS1,O_TS2,...,O_TSN

Nomenclature:
I_TSx(T) = Input neuron for Time Series 'x' at time point 'T'
I_ADDx = Optional additional input neuron 'x' to support the prognosis, not representing a time series
O_TSx = Output neuron for Time Series 'x' at time point 't+1'
n and m define the number of past values which are taken into account and can be defined as script constants separately for each time series within an array in the script. 

When you use additional input neurons to support the prognosis (all values for these additional inputs must known upfront for the whole future extrapolation period!)
then you have to provide the values for these for all future (i.e. extrapolated) data points in a separate CSV file (see string constant 'ADD_EXTRAPOLATION_INPUT_FILE' in the 
header section of the script. 
The script will import them and apply them to the net during extrapolation into the future. The file is only required to be present if you have more input neurons in 
your net than would be needed with respect to the number of output neurons (i.e. input neurons in addition to the time series data themselves).

In each case you should first examine the files which come packed with the example to understand what the input files and the nets have to look like in principle.
Then you might want to go for your own data, create input files and net candidates for it and use the script to perform data extrapolation into the future with MemBrain.

The input CSV file (TimeSeriesInput.csv) in the example represents a mixture of several sine and cosine waves with different frequencies. One of these waves depends on an 
additional input variable 'Month'. You may want to take a look at the original data and its creation in the spread-sheet that comes with the example package. Its file name is 
'TimeSeriesInput.xlsx'. The spread sheet also contains a diagram which depicts the whole time series including the last 10 data points which MemBrain is finally trying
to predict. In the diagram these data points are coloured differently than the others.
Take a look at the tab 'TimeSeriesInput' in the spread sheet. This tab has been exported to the CSV file 'TimeSeriesInput.csv' for the example. Note that it only contains 
the data points up to the beginning of the forecast, i.e. it misses the last 10 data points.
The tab 'AdditionalInputExtrapolation' contains the Month (or rather its cosine representation) for the future 10 data points as additional input to the net. It has been 
exported into the correspondingly named csv file.
 
The script is adjusted to extrapolate 10 points into the future. 
After the script has executed, the result file 'Extrapolated.csv' is automatically opened in your preferred spread sheet program or text editor - depending on your 
corresponding operating systems settings for opening CSV files. 

To perform the same with your data you need to perform actions as described below. Please note that most of the steps below are one-time steps. 
The only thing you need to do for every data set is the creation of corresponding time series input CSV files.

- Prepare at least one net with inputs x(t-n), x(t-n+1)..., x(t), y(t-m), y(t-m+1)..., y(t), ... and outputs x(t+1), y(t+1)... You can select different names of the 
  neurons if you want to.
  The input neurons from left to right must reflect the same meaning as the output neurons from left to right as depicted above!
  If you use additional inputs (as the 'COS(Month)' in the example) then you have to place them TO THE RIGHT of the other input neurons and you will have to 
  create a separate csv file which contains their data points for the extrapolation period as described above. 

Also, don't forget to adjust appropriate normalization ranges for all input and output neurons of your net candidates!

- Copy the files you have created into the directory where the script resides
- In the constant header section of this script adjust the constant values according to your needs.

Note that all your net candidates must have a name that comprises of the value assigned to NET_FILE_NAME_BASE plus a number that starts at '1' for the first net candidate.

For the example, adjust MemBrain as following:
- Edit the teacher 'RPROP' via the Teacher Manager: 
  Disable the option 'Use On-The-Fly Net Error Calculation if possible'. Enable the option 'Reset Net before every Lesson'.
- Select <Teach><Configure Net Error Function>
  On the left hand side select 'Binary by Threshold' and enter the value 0.03
  On the right hand side select 'Mean of Sum'
  
All these are one-time actions only. You can now start the script and it should run seamlessly.

If you want to learn what happens in the script in details try to read through the code. Even if you are not a programmer you should be able to get an 
idea of what is going on according to the comments in the script and the names of the functions that are called. Watch out for the function 'void main()'. 
This is the main program and entry point of the script.
When reading MemBrain script files then best use the Script Debugger that installs together with MemBrain. This brings special syntax highlighting and certainly 
also the possibility to step through the script, inspect variables etc.
*/


// ---------- constants which can be adjusted to tweak the script according to the user's needs -------------------

// The time series input file from which training and validation lesson are generated
const string TIME_SERIES_INPUT_FILE = "TimeSeriesInput.csv";

// Additional input file which is only required if the NN has more inputs than outputs, i.e. 
// if not all input data belongs to time series.
// Note that these addirional inputs must follow on the right of the time series inputs and
// the CSV file referenced here may only contain columns for the additional inputs.
// It must provide as least as many rows as there are future extrapolation points (see EXTRAPOLATE_COUNT)
const string ADD_EXTRAPOLATION_INPUT_FILE = "AdditionalInputsExtrapolation.csv";

// Percentage of data used for validation during evaluating the best net candidate
const double VALIDATION_LESSON_PERCENT = 20.0;

// Number of neural nets to operate on
const uint8 NUMBER_OF_NETS = 5;

// Number of data points to be extrapolated into the future
const uint EXTRAPOLATE_COUNT = 10;

// Common part of the name of all used neural net candidates. To be extended by number of net 1 .. NUMBER_OF_NETS
const string NET_FILE_NAME_BASE = "Net";

// Default training time per net [s]
const uint32 DEFAULT_TRAIN_TIME = 20;

// The name for the generated training lesson file
const string TRAIN_LESSON_NAME = "Train.mbl";
  
// The name for the generated validation lesson file
const string VALIDATE_LESSON_NAME = "Validate.mbl";

// Name of combined lesson used for extrapolation
const string EXTRAPOLATION_LESSON_NAME = "ExtrapolationBase.mbl";

// The name with which the final net is stored to disk
const string FINAL_NET_FILE_NAME = "FinalNet.mbn";

// The teacher to be used
const string TEACHER_NAME = "RPROP";

// The extension appended to a net file name when being saved after training
const string NET_FILE_SAVE_AS_EXTENSION = "_Trained";

// File name for the summary
const string RESULT_SUMMARY_FILE_NAME = "TrainingResultSummary.txt";

// Auto-open the file RESULT_SUMMARY_FILE_NAME in the end of the script?
const bool OPEN_RESULT_SUMMARY = false;

// Name of resulting CSV file that contains all result data
const string RESULT_FILE_NAME = "Extrapolated.csv";

// Auto-open the file RESULT_FILE_NAME in the end of the script?
const bool OPEN_RESULT_FILE = true;

// Field separator
const string LIST_SEPARATOR = ";";

// Decimal separator
const string DEC_SEPARATOR = ",";

// Heading for the additionally generated time column in the file <RESULT_FILE_NAME>
const string TIME_COLUMN_HEADING = "#";

// Start value for the additionally generated time column in the file <RESULT_FILE_NAME>
const double TIME_START_VALUE = 0.0;

// Time step size value for the additionally generated time column in the file <RESULT_FILE_NAME>
const double TIME_STEP_SIZE = 1.0;

// Use MemBrain's feature to capture and reload the best net training state from stock
// after the final training cycle? 
const bool USE_BEST_NET_STOCK_CAPTURE = true;

// Shall MemBrain train the final net with the extrapolation lesson before to extrapolate? 
const bool TRAIN_WITH_EXTRAPOLATION_LESSON = true;

// Shall MemBrain randomize the net again before finally training with the extrapolation lesson?
const bool RANDOMIZE_BEFORE_FINAL_TRAINING = false;

// Naming extension added to the output neurons: Output neurons will be named after the corresponding 
// time series name from the imported CSV file plus this extension.
const string OUTPUT_NAMING_EXTENSION = "(t+1)";

// Defines the number of inputs used for each time series
const uint[] INPUTS_PER_TIME_SERIES =
{
	1
	// For more than one time series add a line for each time series followed by a ','. 
	// Don't place a ',' after the last line!	
};



// ---------- variables with static storage duration below. DON'T CHANGE!!!!!!!!!!!!!! -------------------

// The last known net error for every net
double[] sCurNetErr(NUMBER_OF_NETS);

// The last known minimum net error for every net
double[] sMinNetErr(NUMBER_OF_NETS);

// Index of current net
uint8	sCurNetIdx = 0;

// Training time in seconds
uint32	sTrainTime = 0;

// Has the summary file been opened successfully?
bool sSummaryFileOpen = false;

// The summary file
file sSummaryFile;

// Name of the best net
string sBestNetName;

// To remember the user's settings and restore them after script execution
bool sOldBestNetStockCapture;

// Number of input neurons which are dedicated to time series values 
uint sTimeSeriesInputNeuronCount = 0;


// This is the main script execution function
void main()
{
	// Perform some initialization
	Init();
	
	// Open the first net and open all used windows
	OpenCurrentNet();
	ShowPatternErrorViewer(false);
	ShowErrorViewer(false);
	
	// Create a training and a validation lesson from the time series input CSV file.
	if (!CreateExtrapolationLessonFromTimeSeriesInput())
	{
		AbortScript();
	}
	
	if (!SplitOffTrainingAndValidationLesson())
	{
		AbortScript();
	}
    
    // Request some runtime parameters from the user
	RequestTrainingTime();

	ShowPatternErrorViewer(true);
	ShowErrorViewer(true);
	
	// Train and validate all the nets
	for (uint8 i = 0; i < NUMBER_OF_NETS; i++)
	{
		TrainAndValidateNet(i);
	}

	SummarizeResults();
	
	AnnounceAction("Training with combined lesson and then extrapolating " + EXTRAPOLATE_COUNT + " data points into the future");
	
	// Now train the best net once again with the combined data set and extrapolate
	// into the future (separate script file).
	bool ok = TrainAndExtrapolate(sBestNetName, EXTRAPOLATION_LESSON_NAME, sTrainTime, 
                        		EXTRAPOLATE_COUNT, RESULT_FILE_NAME);
                        		  
    if (ok)
    {
    	Trace("Result data is contained in the file " + RESULT_FILE_NAME + "\r\n");
    	Trace("The file contains two data columns for each time series: The first column contains the input data of the training and the validation lesson"
          	  " (combined into one time series). The second column contains the output data of the net for each input data point.\r\n");
    	if (EXTRAPOLATE_COUNT > 1)
    	{
    		Trace("Note that the last " + (EXTRAPOLATE_COUNT - 1) + " data points in the input column are the output of the net from the last"
          	  	" point in time!\r\n");
        }
          	
        AutoOpenResultFiles();
    }
    else
    {
    	MessageBox("The script could not perform all required tasks.\r\n" 
    			   "An error occurred!"); 
    }
    
    // Cleanup operations after script execution
    CleanUp();
}


// Output a message in the trace window to be the next action of the script
void AnnounceAction(const string &in message)
{
	// Open trace window for sure and leave one empty line after last text entry.
	ShowTraceWin(true);
	Trace(message + "...\n");
}

// Get the file name for the given net index
string GetNetFileFileName(uint8 idx)
{
	return NET_FILE_NAME_BASE + (idx + 1) + ".mbn";
}

// Get the file name for the current net index
string GetNetFileNameCurrentIdx()
{
	return GetNetFileFileName(sCurNetIdx);
}

// Get the file name for the given net ndex to save the trained net as
string GetNetFileNameSaveAs(uint8 idx)
{
	return NET_FILE_NAME_BASE + (idx + 1) + NET_FILE_SAVE_AS_EXTENSION + ".mbn";
}

// Get the file name for the current net ndex to save the trained net as
string GetNetFileNameCurrentIdxSaveAs()
{
	return GetNetFileNameSaveAs(sCurNetIdx);
}


// Open the net corresponding to the current net array index 'sCurNetIdx'
void OpenCurrentNet()
{
	OpenNet(GetNetFileNameCurrentIdx());
}



// Initialize net errors.
void Init()
{
	StopThink();
	StopTeaching();
	
	AnnounceAction("Initializing variables");
	// negative values for net errors indicate invalid values
	for (uint8 i = 0; i < NUMBER_OF_NETS; i++)
	{
		sCurNetErr[i] = -1.0;
		sMinNetErr[i] = -1.0;
	}

	sCurNetIdx = 0;

	AnnounceAction("Selecting teacher");
	SelectTeacher(TEACHER_NAME);
	TeacherSetting(TARGET_ERR, 0.0);
	TeacherSetting(PATTERN_SELECT, ORDERED);
	
	// Remember the user's setting to restore it after script execution
	sOldBestNetStockCapture = GetCaptureBestNetOnStock();
}

// Cleanup operations after script execution
void CleanUp()
{
	// Restore user setting in case we might have modified it
	if (USE_BEST_NET_STOCK_CAPTURE != sOldBestNetStockCapture)
	{
 		SetCaptureBestNetOnStock(sOldBestNetStockCapture);
	}
}


// Request the training time from the user
void RequestTrainingTime()
{
	EDlgRet ret = UserInput("Please enter training time per net in seconds.", DEFAULT_TRAIN_TIME, sTrainTime);

	if (ret != IDOK)
	{
		AbortScript();
	}
}

// Train and validate the net with the given 0-based index
void TrainAndValidateNet(uint8 netIdx)
{
	sCurNetIdx = netIdx;

	string action = "Training and validating net: ";

	action += GetNetFileNameCurrentIdx();
	AnnounceAction(action);

	OpenCurrentNet();
	// Ensure that all nets get the I/O names of the corresponding lesson
	NamesToNet();
	
	RandomizeNet();
	StartTimer(sTrainTime * 1000);

	ETeachResult res = TR_OK;

	StartTeaching();

	// Teach as long as teacher signals OK and as long as timer is not expired.
	// Update net error values during teaching at regular intervals
	while ((res == TR_OK) && !IsTimerExpired() && IsTeaching())
	{
		Sleep(100);
		UpdateNetErrors();

		res = GetLastTeachResult();
	}

	if (IsTeaching())
	{
		// Stop teaching for sure
		StopTeaching();
	}
	UpdateNetErrors();

	SaveNet(GetNetFileNameCurrentIdxSaveAs());
}


// Update the net error variables for later analysis
void UpdateNetErrors()
{
	double err = GetNetError();

	// Store current net error for the given net
	sCurNetErr[sCurNetIdx] = err;

	// Store minimum net error for the given net
	if ((sMinNetErr[sCurNetIdx] < 0) || (err < sMinNetErr[sCurNetIdx]))
	{
		sMinNetErr[sCurNetIdx] = err;
	}
}


// Provide a result summary of the training
void SummarizeResults()
{
	Trace("\n------------------ Training result summary start ----------------------\n\n");

	string result;

	// Try to open result summary file
	sSummaryFileOpen = sSummaryFile.Open(RESULT_SUMMARY_FILE_NAME, FILE_MODE_WRITE | FILE_MODE_CREATE | FILE_SHARE_DENY_NONE);

	if (!sSummaryFileOpen)
	{
		MessageBox("Unable to open summary file!");
	}

	// Create headline in summary file
	Append2SummaryFile("Net" + LIST_SEPARATOR + "Last Net Error" + LIST_SEPARATOR + "Min Net Error\n");

	double minNetErr = -1;
	uint8 bestNetIdx = 0;

	for (uint8 i = 0; i < NUMBER_OF_NETS; i++)
	{
		result = GetNetFileNameSaveAs(i);
		Trace(result + ":\n");
		Append2SummaryFile(result + LIST_SEPARATOR);

		result = sCurNetErr[i];
		result = ConvertDoubleStringToLocale(result);
		Trace("Last known net error: " + result + ":\n");
		Append2SummaryFile(result + LIST_SEPARATOR);

		result = sMinNetErr[i];
		result = ConvertDoubleStringToLocale(result);
		Trace("Minimum net error: " + result + ":\n\n");
		Append2SummaryFile(result + "\n");

		if ((minNetErr < 0) || (sCurNetErr[i] < minNetErr))
		{
			minNetErr = sCurNetErr[i];
			bestNetIdx = i;
		}
	}

	sBestNetName = GetNetFileNameSaveAs(bestNetIdx);
	Trace("The best trained net is '" + sBestNetName + "'\n");

	Trace("\n------------------ Training result summary end ----------------------\n\n");

	Append2SummaryFile("\nBest trained net: " + sBestNetName + "\n");

	if (sSummaryFileOpen)
	{
		Trace("All results have been additionally summarized in the file '" + RESULT_SUMMARY_FILE_NAME + "'.\n");
		sSummaryFile.Close();
	}

}


// Append text to the summary file if open
void Append2SummaryFile(string text)
{
	if (sSummaryFileOpen)
	{
		if (!sSummaryFile.Write(text))
		{
			MessageBox("Error in writing to summary file!");
			sSummaryFile.Close();
			sSummaryFileOpen = false;
		}
	}
}


// Convert the given number string to incorporate local decimal separator
const string ConvertDoubleStringToLocale(const string &in numStr)
{
	if (DEC_SEPARATOR == ".")
	{ // Nothing to do if anglo-american notation is adjusted
		return numStr;
	}

	int idx = numStr.findFirst(".", 0);

	if (idx < 0)
	{ // Nothing to do if string does not contain separator
		return numStr;
	}

	// The string contains a separator. Replace by user defined separator now...

	// copy part left from separator
	string ret = numStr.substr(0, idx);

	// Add user defined separator
	ret += DEC_SEPARATOR;

	// copy part right from separator
	ret += numStr.substr(idx + 1, numStr.length() - idx - 1);

	return ret;
}

// Perform a left-shift action on all input neuron activities for a given time series.
// Afterwards set activation of input neuron for the last time point in the time series to 
// the given new value 
bool ShiftInputTimeSeriesToLeftInNet(uint timeSeriesNum, double newNewestValue)
{
	double act;
	uint neuronNum = 1;
	uint numberTimePoints = INPUTS_PER_TIME_SERIES[timeSeriesNum - 1];
	bool ok = true;
	
	for (uint i = 0; i < timeSeriesNum - 1; i++)
	{
		neuronNum += INPUTS_PER_TIME_SERIES[i];
	}
	
	for (uint i = 0; i < numberTimePoints - 1; i++)
	{
		ok = ok && GetInputAct(neuronNum + 1, act);
		ok = ok && ApplyInputAct(neuronNum, act);
		neuronNum++;
	}
	ok = ok && ApplyInputAct(neuronNum, newNewestValue);
	
	return ok;
}


// Final training and extrapolation extrapolate into the future
// Return false on error
bool TrainAndExtrapolate(string &in netFile, string &in lessonFile, uint trainTime, 
                        uint extrapolateCount, string &in resultLessonName)
{
    // Open net, delete all lessons except for #1, create an additional, empty lesson #2
    // import training data into lesson #1
    OpenNet(netFile);
    SetLessonCount(1);
    // Load training lesson
    SelectLesson(1);
    LoadLesson(lessonFile);
    // The net error shall be calculated based on the training lesson
    SelectNetErrLesson(1);
    
    if (TRAIN_WITH_EXTRAPOLATION_LESSON)
    {
    	// Train and save the net
    	TrainForExtrapolation(trainTime);
    }
    SaveNet(FINAL_NET_FILE_NAME);
    
    // Does the net have more inputs as there are time series (i.e. additional inputs)?
    // During further extrapolation we need to fetch values for these inputs from an additional 
    // CSV file.
    uint inCount = GetInputCount();
    uint timeSeriesCount = GetOutputCount();
    uint addInputsCount = 0;
    
    if (inCount > sTimeSeriesInputNeuronCount)
    {
    	addInputsCount = inCount - sTimeSeriesInputNeuronCount;
    	// Create lesson 2 and 3 newly (empty)
    	SetLessonCount(3);
    	// Inport the additional input values into lesson #3
    	SelectLesson(3);
    	SetLessonInputCount(addInputsCount);
    	SetLessonOutputCount(0);
    	if (EXTRAPOLATE_COUNT > 0)
    	{
    		ImportLessonInputsRaw(ADD_EXTRAPOLATION_INPUT_FILE);
    		if (GetLessonSize() < EXTRAPOLATE_COUNT)
    		{
    			Trace("Error: Not enough input lines in file " + ADD_EXTRAPOLATION_INPUT_FILE + "\r\n" +
    			  	"to extrapolate " + EXTRAPOLATE_COUNT + " into the future!\r\n");
    			AbortScript();
    		}
    	}
    }
    else
    {
    	// Create lesson 2 newly (empty)
    	SetLessonCount(2);
    }
    
    SelectLesson(1);
    // Start recording of output activation values into lesson #2
    SetRecordingType(RT_ACT);
    StartRecording(2);
    EnableLessonEditorUpdate(false);
    // Put net into a well defined state (time variant network!)
    ResetNet();
    
    // Disable view update during think to increase speed
    ViewSetting(UPDATE_THINK, false);
    
    // Let training lesson run one time through the net
    ThinkLesson();
    SleepExec();
    // Update lesson editor once
    EnableLessonEditorUpdate(true);
    EnableLessonEditorUpdate(false);
    
    // Up to now the output reaction of the net on the input of the currently known time series 
    // has been recorded.
    // From now on the net will calculate future data points. The first future data point is based on the 
    // last known target output values of the lesson since this is the last known data point from the past
    // (i.e. the 'now' data point). The script will now apply this last known data point to the inputs of the 
    // net and perform one single think step. This gives us the first future (i.e. extrapolated) data point.
    // For all following future data points the script will read the outputs of the net, apply them to the inputs 
    // and perform a think step.
    // All results will still be recorded into the result lesson since recording is still active.
    // Note that the order of the inputs from left to right must be exactly the same as the order of the outputs 
    // and if there are additional input neurons then they must be positioned to the right of the time dependend
    // inputs!
    
    ShowPatternErrorViewer(false);
    
    bool ok = true;
    
    if (extrapolateCount > 0)
    {
        Trace("Extrapolating into the future...\r\n");
        // Select the last pattern in the lesson from which we need the output values in order to 
        // apply them to the input of the net.
        SelectPattern(GetLessonSize());
        double act;
        inCount = GetInputCount();
        uint outCount = GetOutputCount();
        // We need at least as many inputs as we have outputs in order to feed back all outputs to the inputs
        // and we need at least as many inputs as we have defined to be there for past time series values
        ok = (inCount >= outCount) && (inCount >= sTimeSeriesInputNeuronCount);
        
        for (uint i = 1; i <= outCount; i++)
        {
        	// Get the lesson output and apply it to the net's input.
        	ok = ok && GetPatternOutput(i, act);
        	ok = ok && ShiftInputTimeSeriesToLeftInNet(i, act);
        }
                
        // Perform the think step.
        if (ok)
        {
            ThinkSteps(1);
            SleepExec();
        }
                
        // The first future data point has been acquired. Now go for the next ones using 
        // the current output of the net as next input.
        uint extCount = 1;
        
        for (uint j = 0; ok && (j < extrapolateCount - 1); j++)
        {
            for (uint i = 1; ok && (i <= outCount); i++)
        	{
        		// Get the lesson output and apply it to the net's input.
        		ok = ok && GetOutputAct(i, act);
        		ok = ok && ShiftInputTimeSeriesToLeftInNet(i, act);
        	}
        	// Also apply additional inputs if there are any
        	if (addInputsCount > 0)
        	{
        		SelectLesson(3);
        		SelectPattern(j + 1);
        		for (uint addIn = 1; addIn <= addInputsCount; addIn++)
        		{
         			ok = ok && GetPatternInput(addIn, act);
        			ok = ok && ApplyInputAct(sTimeSeriesInputNeuronCount + addIn, act);
        		}
        	}
            ThinkSteps(1);
            SleepExec();
            extCount++;
            if ((extCount % 100) == 0)
    		{
    			Trace(extCount + " of " + extrapolateCount + " future data points processed\r\n");
    		} 
        } 
        
        Trace(extrapolateCount + " of " + extrapolateCount + " future data points processed\r\n");
       
        if (!ok)
        {
            MessageBox("Error!");
        }
    }
    else
    {
        Trace("No future data points to extrapolate!\r\n");
    }
    
    // Stop recording, select lesson #2 and export to csv.
    StopRecording();
    EnableLessonEditorUpdate(true);
    
    // Re-enable view update during think
    ViewSetting(UPDATE_THINK, true);
                      
    SelectLesson(2);
    
    if (ok)
    {
    	// Add a time column to the left of the inputs to simplify post-processing 
    	// in external programs
    	AddTimeColumn(2);
    	// Export into result CSV file
    	AnnounceAction("Export to CSV");
    	ExportLessonRaw(resultLessonName);
    }
    
    return ok;
}


// Train the net for extrapolation.
void TrainForExtrapolation(uint trainTime)
{
    if (RANDOMIZE_BEFORE_FINAL_TRAINING)
    {
    	RandomizeNet();
    }
    
    // Adjust the option to capture best net during training on stock
    SetCaptureBestNetOnStock(USE_BEST_NET_STOCK_CAPTURE);
    
    StartTimer(trainTime * 1000);

	ETeachResult res = TR_OK;

	StartTeaching();

	// Teach as long as teacher signals OK and as long as timer is not expired.
	// Update net error values during teaching at regular intervals
	while ((res == TR_OK) && !IsTimerExpired() && IsTeaching())
	{
		Sleep(100);
		
        res = GetLastTeachResult();
	}

	if (IsTeaching())
	{
		// Stop teaching for sure
		StopTeaching();
	}
	
	// Reload the best training version of this net from stock if desired
	if (USE_BEST_NET_STOCK_CAPTURE)
	{
		string fileName;
		GetCurrentNetFileName(fileName);
		uint slot = GetLastTeacherAutoCaptureStockSlot();
		string message = "Loading best training version of net ";
		message += fileName + "\r\n";
		message += "from stock slot ";
		message += slot + "...\r\n";
		AnnounceAction(message);
		LoadNetFromStock(slot);
	}
}

// Add a time column to the left in the specified lesson number
void AddTimeColumn(uint lessonNum)
{
	Trace("Add time column to recorded data...\r\n");
	// Select the recorded lesson
	SelectLesson(lessonNum);
	
	// Disable Lesson Editor screen update for higher performance
	EnableLessonEditorUpdate(false);
	
	// We need one more input column. This will by default be generated to the right of the other columns
    uint oldInCount = GetLessonInputCount();
    SetLessonInputCount(oldInCount + 1);
    // Since we want our time column on the left hand side we need to move all columns in the input section 
    // one position to the right, including the names of the columns
    
    // First move all names
    string name;
    for (uint col = oldInCount; col >= 1; col--)
    {
    	GetLessonInputName(col, name);
    	SetLessonInputName(col + 1, name);
    }  
    
    // Now move all data rows
    uint rows = GetLessonSize();
    double value;
    Trace("Right-shift recorded data...\r\n");
    for (uint row = 1; row <= rows; row++)
    {
    	SelectPattern(row);
    	for (uint col = oldInCount; col >= 1; col--)
    	{
    		GetPatternInput(col, value);
    		SetPatternInput(col + 1, value);
    	} 
    	if ((row % 100) == 0)
    	{
    		Trace(row + " of " + rows + " patterns processed\r\n");
    	} 
    }
    Trace(rows + " of " + rows + " patterns processed\r\n");   
    
    
    // Set the heading of the new column on the left side
    SetLessonInputName(1, TIME_COLUMN_HEADING);
    // Generate time axis data
    double time = TIME_START_VALUE;
    
    Trace("Insert time axis values...\r\n");
    for (uint row = 1; row <= rows; row++)
    {
    	SelectPattern(row);
    	SetPatternInput(1, time);
    	time += TIME_STEP_SIZE;	
    	if ((row % 100) == 0)
    	{
    		Trace(row + " of " + rows + " patterns processed\r\n");
    	}
    }
    Trace(rows + " of " + rows + " patterns processed\r\n");  
	
	EnableLessonEditorUpdate(true);
}


// Automatically open the result files as configured in the script constants section
void AutoOpenResultFiles()
{
	if (OPEN_RESULT_SUMMARY)
	{
		ShellExecute("open", RESULT_SUMMARY_FILE_NAME, "");
	}
	
	if (OPEN_RESULT_FILE)
	{
		ShellExecute("open", RESULT_FILE_NAME, "");
	}
}

// Create a combined extrapolation lesson from the time series input CSV file.
// Data from time points t-n, t-n+1, ... will be combined with data from time points t+1 
// to build input data / expected output data.
bool CreateExtrapolationLessonFromTimeSeriesInput()
{
	AnnounceAction("Create extrapolation lesson\n");
	// Import the time series input file into lesson #1 according to the format of
 	// the loaded net
 	SetLessonCount(1);
 	ClearLesson();
	NamesFromNet();
	// The number of different time series to be trained is reflected in the loaded net
	// by the number of outputs 
    uint timeSeriesCount = GetOutputCount();
    uint inCount = GetInputCount();
    uint timeSeriesLength = 0;
    double value;
    uint maxInputsPerTimeSeries = 1;
    
    // Disable Lesson Editor screen update for higher performance
	EnableLessonEditorUpdate(false);
	    
    // The net must have the correct number of inputs according to the number of used data points
    // for each time series and optional additional input data columns.
    // Note that all additional inputs must be placed to the right of the time series inputs!
        
    // First, we need a definition of the number of used inputs for each time series 
    uint temp = INPUTS_PER_TIME_SERIES.length();
    bool ok = (timeSeriesCount == INPUTS_PER_TIME_SERIES.length());
    
    if (ok)
    {
    	for (uint i = 0; i < timeSeriesCount; i++)
    	{
    		uint inputs = INPUTS_PER_TIME_SERIES[i];
    		sTimeSeriesInputNeuronCount += inputs;
    		if (inputs > maxInputsPerTimeSeries)
    		{
    			maxInputsPerTimeSeries = inputs;
    		}
    	}
    	
    	ok = (inCount >= sTimeSeriesInputNeuronCount);
    	if (ok)
    	{
    		uint addInputCount = inCount - sTimeSeriesInputNeuronCount;
    		// In order to correctly import the time series input file we need as many columns as there
    		// are time series plus the number of additional input columns which are not time series.
    		SetLessonInputCount(timeSeriesCount + addInputCount);
    		// The time series input file does not have any output columns 
    		ImportLessonInputsRaw(TIME_SERIES_INPUT_FILE);
    		SaveLesson("InputLesson.mbl");
			// Determine length of time series (number of data points per time serie)
			timeSeriesLength = GetLessonSize();
			// A time series must consist of at least two data points and of at least enough data points to
			// for one valid data pattern
			ok = (timeSeriesLength > maxInputsPerTimeSeries);
			if (ok)
			{
				SetLessonCount(2);
				uint targetLessonSize = timeSeriesLength - maxInputsPerTimeSeries;
				SetLessonSize(2, targetLessonSize, "InputLesson.mbl");
				CreateExtrapolationLesson(1, 2, maxInputsPerTimeSeries, sTimeSeriesInputNeuronCount);
			}
			else
    		{
    			Trace("Input time series must contain at least " + maxInputsPerTimeSeries + " data rows!\r\n");
    		}
		}
		else
		{
			Trace("At least " + sTimeSeriesInputNeuronCount + " input neurons expected according to the number of outputs and the \r\n"
			      "definition of the script constant 'INPUTS_PER_TIME_SERIES'\r\n"); 
		}
    }
    else
    {
    	Trace("The script array 'INPUTS_PER_TIME_SERIES' must provide exactly one value for each time series\r\n"
    		 "(i.e. one for each output neuron)!\r\n");
    }
    
    EnableLessonEditorUpdate(true);
    
	return ok;
}

// Create a combined extrapolation lesson from the time series input CSV file.
// Time series data is already imported into lesson <sourceLessonNum>.
// The target lesson <targetLessonNum> has already been allocated.
// Data from time points t-n, t-n+1, ... will be combined with data from time points t+1 
// to build input data / expected output data.
void CreateExtrapolationLesson(uint sourceLessonNum, uint targetLessonNum, uint maxInputsPerTimeSeries, uint timeSeriesInputCount)
{
	SelectLesson(sourceLessonNum);
	
	uint timeSeriesLength = GetLessonSize();
	uint sourceLessonInCount = GetLessonInputCount();
	uint timeSeriesCount = GetOutputCount();
		
	SelectLesson(targetLessonNum);
	NamesFromNet();
	
	// Create meaningful names for the outputs according to the inputs
    string name;
    uint inputPos = 1;
    
    EnableLessonEditorUpdate(false);
    
    for (uint col = 1; col <= timeSeriesCount; col++)
    {
    	SelectLesson(sourceLessonNum);
    	GetLessonInputName(col, name);
    	SelectLesson(targetLessonNum);
    	SetLessonOutputName(col, name + OUTPUT_NAMING_EXTENSION);
    	
    	uint timePointCount = INPUTS_PER_TIME_SERIES[col - 1];
    	
    	for (uint i = timePointCount; i > 0; i--)
    	{
    		if (i == 1)
    		{
    			SetLessonInputName(inputPos, name + "(t)");
    		}
    		else
    		{
    			SetLessonInputName(inputPos, name + "(t-" + (i - 1) + ")");
    		}
    		inputPos++;
    	}
    }
    
    EnableLessonEditorUpdate(true);
    // Assign the lesson names to the net
    NamesToNet();
    EnableLessonEditorUpdate(false);
    
    // Now perform the time shift: For each pattern (i.e. input data row) copy 
	// the corresponding data fields from the next row as expected output data 
	// (data point t+1 will become the expected output of data points t).
	// For the last input row there is no expected output. We have to delete this input
	// row.
	double value;
	uint sourceRow = 1;
	double[] sourceInputVector;
			
	sourceInputVector.resize(sourceLessonInCount);
		
	// Pre-fill all inputs with past values and generate first training pattern once all 
	// input values are filled.
	// Create and first pattern in the target lesson
    SelectLesson(targetLessonNum);
    
    for (sourceRow = 1; sourceRow <= timeSeriesLength; sourceRow++)
    {
    	SelectLesson(sourceLessonNum);
    	SelectPattern(sourceRow);
    	
    	// Read in whole input row to simplify processing later on
    	for (uint inputNum = 1; inputNum <= sourceLessonInCount; inputNum++)
    	{
    		GetPatternInput(inputNum, value);
    		sourceInputVector[inputNum - 1] = value;
    	}
    	// Select target pattern
    	SelectLesson(targetLessonNum);
        SelectPattern(1);
    	
    	if (sourceRow > maxInputsPerTimeSeries)
    	{ // All inputs pre-filled. Add output values from next input lesson pattern now and exit this preparation 
    	  // loop
    		for (uint inputNum = 1; inputNum <= timeSeriesCount; inputNum++)
    		{
    			SetPatternOutput(inputNum, sourceInputVector[inputNum - 1]);
    		}
    		break;
    	}
    	
    	for (uint inputNum = 1; inputNum <= sourceLessonInCount; inputNum++)
    	{
    		if (inputNum > timeSeriesCount)
    		{ // Auxiliary input: Just copy to target lesson as inputs to the right of the time series inputs
    			uint targetCol = timeSeriesInputCount + (inputNum - timeSeriesCount);
    			SetPatternInput(targetCol, sourceInputVector[inputNum - 1]);	
    		} 
    		else
    		{ // Time series input
    			uint pastPoints = INPUTS_PER_TIME_SERIES[inputNum - 1];
    			if (sourceRow > maxInputsPerTimeSeries - pastPoints)
    			{
    				uint targetCol = 1;
    				for (uint i = 0; i < inputNum - 1; i++)
    				{
    					targetCol += INPUTS_PER_TIME_SERIES[i];
    				}
    				targetCol += (sourceRow - (maxInputsPerTimeSeries - pastPoints) - 1); 
    				SetPatternInput(targetCol, sourceInputVector[inputNum - 1]);
    			} 
    		}
    	}	
    }
    // First pattern created now.
    sourceRow++;
    uint targetPattern = 1;
    // Create all following patterns now using time shift within each time series input field:
    // 1.) Shift inputs to the left by one
    // 2.) Move pattern outputs to the inputs (time point t, right-most of each input field)
    // 3.) Take input values from time series input and set them as pattern outputs
    double[] targetTimeSeriesInputVector;
    double[] targetTimeSeriesOutputVector;
    			
	targetTimeSeriesInputVector.resize(timeSeriesInputCount);
	targetTimeSeriesOutputVector.resize(timeSeriesCount);
	    
    for (; sourceRow <= timeSeriesLength; sourceRow++)
    {
    	if ((sourceRow % 100) == 0)
    	{
    		Trace(sourceRow + " of " + timeSeriesLength + " patterns processed\r\n");
    	}
    	SelectLesson(sourceLessonNum);
    	SelectPattern(sourceRow);
    	// Read in whole input row to simplify processing later on
    	for (uint inputNum = 1; inputNum <= sourceLessonInCount; inputNum++)
    	{
    		GetPatternInput(inputNum, value);
    		sourceInputVector[inputNum - 1] = value;
    	}
    	SelectLesson(targetLessonNum);
    	// Read in all target lesson time series inputs to simplify processing later on
    	for (uint inputNum = 1; inputNum <= timeSeriesInputCount; inputNum++)
    	{
    		GetPatternInput(inputNum, value);
    		targetTimeSeriesInputVector[inputNum - 1] = value;
    	}
    	// Read in all target lesson time series outputs to simplify processing later on
    	for (uint outputNum = 1; outputNum <= timeSeriesCount; outputNum++)
    	{
    		GetPatternOutput(outputNum, value);
    		targetTimeSeriesOutputVector[outputNum - 1] = value;
    	}
    	
    	targetPattern++;
    	SelectPattern(targetPattern);
    	
    	for (uint inputNum = 1; inputNum <= sourceLessonInCount; inputNum++)
    	{
    		if (inputNum > timeSeriesCount)
    		{ // Auxiliary input: Just copy to target lesson as inputs to the right of the time series inputs
    			uint targetCol = timeSeriesInputCount + (inputNum - timeSeriesCount);
    			SetPatternInput(targetCol, sourceInputVector[inputNum - 1]);	
    		} 
    		else
    		{ // Time series input
    			uint offset = 1;
    			for (uint i = 0; i < inputNum - 1; i++)
    			{
    				offset += INPUTS_PER_TIME_SERIES[i];
    			}
    			uint pastPoints = INPUTS_PER_TIME_SERIES[inputNum - 1];
    			for (uint i = 0; i < pastPoints - 1; i++)
    			{ // Time shift to the left
    				value = targetTimeSeriesInputVector[offset + i];
    				SetPatternInput(offset + i, value);
    			} 
    			// Output copied to the last input time slot
    			value = targetTimeSeriesOutputVector[inputNum - 1];
    			SetPatternInput(offset + pastPoints - 1, value);
    		}
    	}
    	
    	// Add output values from input lesson pattern now
    	for (uint inputNum = 1; inputNum <= timeSeriesCount; inputNum++)
    	{
    		SetPatternOutput(inputNum, sourceInputVector[inputNum - 1]);
    	}	
    }
    Trace(timeSeriesLength + " of " + timeSeriesLength + " patterns processed\r\n");
        
    EnableLessonEditorUpdate(true);
    
    // The just created lesson with all data points combined of t and t+1 is our
    // extrapolation lesson used in the end of the script processing for final
    // training and extrapolating into the future. Save it to file now.
    SaveLesson(EXTRAPOLATION_LESSON_NAME);
}


// Split off training and validation lesson from lesson #2 which already contains
// the extrapolation lesson 
bool SplitOffTrainingAndValidationLesson()
{
	// Disable Lesson Editor screen update for higher performance
	EnableLessonEditorUpdate(false);
	
	uint timeSeriesLength = GetLessonSize();
	uint valiLength = uint(VALIDATION_LESSON_PERCENT * timeSeriesLength / 100); 
	
	bool ok = (valiLength > 0) && (valiLength < timeSeriesLength);
	
	if (ok)
	{
		// First delete all validation pattern from lesson #2 and save it as training lesson
		SelectLesson(2);
		uint size = GetLessonSize();
		Trace("Splitting off validation lesson...\r\n");
    	for (uint i = 1; i <= valiLength; i++)
    	{
    		// Deleting from the end of the lesson is faster than deleting in between
    		SelectPattern(size);
    		DeletePattern();
    		size--;
    		if ((i % 100) == 0)
    		{
    			Trace(i + " of " + valiLength + " patterns processed\r\n");
    		}
    	}
    	Trace(valiLength + " of " + valiLength + " patterns processed\r\n");
    	SaveLesson(TRAIN_LESSON_NAME);
    	
    	uint trainLen = GetLessonSize();
    	
    	// Create new empty lesson. Delete all patterns which belong to the training
    	// lesson and save as validation lesson
		SetLessonCount(3);
		SelectLesson(3);
		LoadLesson(EXTRAPOLATION_LESSON_NAME);
		AnnounceAction("Splitting off training lesson...\r\n");
		for (uint i = 1; i <= trainLen; i++)
    	{
    		SelectPattern(1);
    		DeletePattern();
    		if ((i % 100) == 0)
    		{
    			Trace(i + " of " + trainLen + " patterns processed\r\n");
    		}
    	}
    	Trace(trainLen + " of " + trainLen + " patterns processed\r\n");
    	
    	SaveLesson(VALIDATE_LESSON_NAME);
	}
	else
	{
    	Trace("Not enough data points to create validation lesson!\r\n");
    }
    
    EnableLessonEditorUpdate(true);
    
    if (ok)
    {
    	SelectLesson(2);
    	SelectNetErrLesson(3);
    }
	
	return ok;
}

/*
	Set the size of the given lesson. Optionally, use a a helper lesson in order to speed up the
	lesson resizing process compared to pure operation by AddPattern(): For large lessons this is very slow.
	If the parameter 'helperLessonFileName' is empty then no helper lesson is used. 
*/
void SetLessonSize(uint lessonNum, uint targetSize, string &in helperLessonFileName)
{
	SelectLesson(lessonNum);
	if (helperLessonFileName != "")
	{
		LoadLesson(helperLessonFileName);
	}
	uint size = GetLessonSize();
	int delta = targetSize - size;
	
	if (delta > 0)
	{
		for (int i = 0; i < delta; i++)
		{
			AddPattern();
		}
	}
	else if (delta < 0)
	{
		delta = -delta;
		for (int i = 0; i < delta; i++)
		{
			SelectPattern(size - i);
			DeletePattern();
		}
	}
}



// Automatically called when script aborts (e.g. error)
void OnAbortScript()
{
	StopTeaching();
	//MessageBox("The script was aborted before having finished!");
}
